{% for decorator in decorators -%}
{{ decorator }}
{% endfor -%}
class {{ class_name }}({{ base_class }}):
{%- if description %}
    """
    {{ description }}
    """
{%- endif %}
{%- for field in fields %}
    {{ field.name }} = {{ field.default }}
{%- endfor -%}
